package cc.smtweb.system.bpm.web.engine.flow;

import cc.smtweb.framework.core.common.SwMap;
import cc.smtweb.framework.core.db.DbEngine;
import cc.smtweb.framework.core.db.EntityDao;
import cc.smtweb.framework.core.db.EntityHelper;
import cc.smtweb.framework.core.exception.BizException;
import cc.smtweb.framework.core.exception.SwException;
import cc.smtweb.framework.core.session.UserSession;
import cc.smtweb.framework.core.util.CommUtil;
import cc.smtweb.framework.core.util.DateUtil;
import cc.smtweb.framework.core.util.NumberUtil;
import cc.smtweb.system.bpm.web.design.flow.FlowConst;
import cc.smtweb.system.bpm.web.design.flow.ModelProc;
import cc.smtweb.system.bpm.web.design.flow.ModelProcHelper;
import cc.smtweb.system.bpm.web.design.flow.define.Activity;
import cc.smtweb.system.bpm.web.design.flow.define.ProcInfo;
import cc.smtweb.system.bpm.web.design.form.ModelForm;
import cc.smtweb.system.bpm.web.design.form.ModelFormHelper;
import cc.smtweb.system.bpm.web.engine.flow.entity.*;
import cc.smtweb.system.bpm.web.sys.user.dept.Dept;
import cc.smtweb.system.bpm.web.sys.user.dept.DeptCache;
import cc.smtweb.system.bpm.web.sys.user.party.Party;
import cc.smtweb.system.bpm.web.sys.user.party.PartyCache;
import cc.smtweb.system.bpm.web.sys.user.user.UserCache;
import cc.smtweb.system.bpm.web.sys.user.userGroup.UserGroupHelper;
import lombok.Data;
import org.apache.commons.lang3.StringUtils;
import org.springframework.lang.NonNull;

import java.util.*;

/**
 * Created by Akmm at 2020/3/4 11:35
 * 单据流程实例组件
 */
@Data
public class FlowInstance {
    //会话信息
    private UserSession us;
    //流程定义
    private ModelProc procDef;
    //流程实例
    private ProcInst procInst;

    //活动任务
    private Task actInst;

    //操作类型 新增/编辑/查看
    private int opt_mode = FlowConst.OperatorType.VIEW;

    //流程执行变量
//    public Map<String, String> variables = new HashMap<>();

    //当前任务
    private List<Map<String, String>> tasks = new ArrayList<>();
    //数据提供者
    private FlowProvider provider = new FlowProvider();
    
    private FlowListener listener;

    public FlowInstance(@NonNull FlowListener listener, UserSession us) {
        this.listener = listener;
        this.us = us;
    }

    //能否编辑
    public boolean canEdit() {
        //新增状态可以编辑
        if (opt_mode == FlowConst.OperatorType.NEW || procInst.getStatu() == FlowConst.InstanceStatu.BEGIN.value)
            return true;
        return false;
    }


    /**
     * 流程条件
     *
     * @param expr
     * @return
     */
    public boolean execTrans(String expr) {
        if (StringUtils.isBlank(expr)) return true;
        SwMap params = new SwMap();
        params.putAll(procInst.getData());
        listener.setExprParam(params);
        Object o = NumberUtil.calcExprMapObject(expr, params);
        return o != null && (Boolean) o;
    }

    /**
     * 新建流程实例
     *
     * @param bill_type
     * @throws Exception
     */
    public void create(long bill_type) {
        Party loginParty = PartyCache.getInstance().get(us.getPartyId());
        if (EntityHelper.isEmpty(loginParty)) {
            throw new BizException("当前登录人员管理单位信息为空，请检查后保存！");
        }

        Dept loginDept = DeptCache.getInstance().get(us.getDeptId());
        if (EntityHelper.isEmpty(loginDept)) {
            loginDept = new Dept();
        }

        procDef = ModelProcHelper.getBillProc(bill_type, us.getUserId()); //根据单据分类获取流程定义
        if (EntityHelper.isEmpty(procDef)) {
//            throw new BizException("没有找到对应的流程定义，请到【系统设置】-【单据流程分配】为当前单据类型【" + BillTypeCache.getInstance().getName(bill_type) + "】分配工作流程！");
        }

        Activity start = ModelProcHelper.getFirstActivity(procDef.getEntityId());
        if (start == null) {
            throw new BizException("对应的流程没有定义有效的活动步骤！");
        }

        opt_mode = FlowConst.OperatorType.NEW;

        procInst = new ProcInst();
        procInst.init();
        procInst.setEntityId(DbEngine.getInstance().nextId());
        procInst.setPartyId(loginParty.getEntityId());
        procInst.setDeptId(loginDept.getEntityId());
        procInst.setBillType(bill_type);
        procInst.setMakeDate(DateUtil.nowDateTimeLong());
        procInst.setPrcId(procDef.getEntityId());
        procInst.setStartTime(DateUtil.nowDateTimeLong());
        procInst.setUserId(us.getUserId());
        procInst.setStatu(FlowConst.InstanceStatu.BEGIN.value);

        actInst = new Task();
        actInst.init();
        actInst.setEntityId(DbEngine.getInstance().nextId());
        actInst.setId(procInst.getEntityId());
        actInst.setPrcId(procDef.getEntityId());
        actInst.setActId(start.getId());
        actInst.setHandler(us.getUserId());
        actInst.setStartTime(DateUtil.nowDateTimeLong());
        actInst.setStatu(FlowConst.ActivityStatu.HANDLE.value);
        actInst.setMake(true);
        actInst.setSign(false);
        actInst.setReject(false);
        actInst.setRetake(false);

        procInst.setTaskId(actInst.getEntityId());
    }

    /**
     * 办理
     *
     * @throws Exception
     */
    public void handler() {
        long user_id = us.getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
//        checkPermission();
        if (actInst.getStatu() != FlowConst.ActivityStatu.WAIT.value) {
            throw new BizException("当前任务非待办状态，不能办理失败！");
        }

        actInst.setHandler(user_id);
        actInst.setStartTime(DateUtil.nowDateTimeLong());
        actInst.setStatu(FlowConst.ActivityStatu.HANDLE.value);

        procInst.setTaskId(actInst.getEntityId());
        procInst.setStatu(FlowConst.InstanceStatu.RUNING.value);

        getTaskDao().updateEntity(actInst);
        getProcInstDao().updateEntity(procInst);

        buildBillLog(FlowConst.Button.HANDLE.value, "-", actInst, null);
        listener.handle(actInst);
    }

    //从页面参数加载流程实例
    public void readFromPage(SwMap map) {
        procInst.readFromMap(map);
        actInst.setPriId(procInst.getId());
        actInst.setEntityId(procInst.getTaskId());
    }

    /**
     * 从数据库加载已有流程实例
     *
     * @param bill_id
     * @throws Exception
     */
    public void load(long bill_id, long task_id) {
        procInst = provider.findProcInst(bill_id);
        procDef = ModelProcHelper.getFromCache(procInst.getPrcId());
        task_id = getTaskId(task_id);
        if (task_id > 0L) {
            actInst = provider.findTask(task_id);
        }
    }

    //加载审批意见
    public List<Comment> loadComment() {
        return provider.findComment(procInst.getId());
    }

    /**
     * 当前任务处于待办时：
     * 上一步办理人进来，看上一个步骤，否则看当前步骤
     *
     * @param task_id
     * @return
     */
    private long getTaskId(long task_id) {
        if (task_id > 0L) return task_id;
        Task curTask = provider.findTask(task_id);
        if (EntityHelper.isEmpty(curTask) || curTask.getStatu() != FlowConst.ActivityStatu.WAIT.value) return task_id;
        List<Task> preActs = provider.findBeforeTasks(task_id);
        if (CommUtil.isEmpty(preActs)) return task_id;
        for (Task act : preActs) {
            if (act.getHandler() == us.getUserId()) {
                return act.getId();
            }
        }
        return task_id;
    }

    /**
     * 矫正默认任务及获取可选任务列表，供界面选择
     *
     * @return
     * @throws Exception
     */
    public List<Task> adjustTasks() {
        List<Task> listTask = new ArrayList<>();
        if (opt_mode == FlowConst.OperatorType.NEW || procInst.getStatu() == FlowConst.InstanceStatu.DISUSE.value) {
            listTask.add(actInst);
            return listTask;
        }
        //获取当前单据活动任务列表及当前任务
//        act_inst = null;
        List<Task> list = provider.findCurTask(procInst.getId());
        listTask.addAll(list);
        //当前登录用户
        String sqlIn = "";
        for (Task act : list) {
            //如果办理人是当前用户，则置此任务为当前任务
//            if (act_inst == null) {
            //我办理中的任务：非待办，且是我的任务
            if (act.getStatu() != FlowConst.ActivityStatu.WAIT.value) {
                if (us.getUserId() == act.getHandler()) {
                    actInst = act;
                }
            } else if (FlowHelper.isMyWaitedTask(act.getId(), us.getUserId())) {
                //我的待办任务
                actInst = act;
                act.setHandler(us.getUserId());
            } else {
                Activity ae = procDef.getProcInfo().findActivity(act.getActId());
                if (ae != null && ae.getType() == FlowConst.ActivityType.END.value) {
                    actInst = act;
                }
            }
//            }
            //如果当前任务为待办理，那么要找其前置任务，作为活动任务，方便取回
            if (act.getStatu() == FlowConst.ActivityStatu.WAIT.value) {
                sqlIn += "," + act.getEntityId();
            }
        }
        //找待办任务的前置任务，便于取回
        if (StringUtils.isNotEmpty(sqlIn)) {
            List<Task> list1 = provider.findBeforeTasks(sqlIn.substring(1));
            listTask.addAll(list1);
            if (actInst == null) {
                for (Task act : list1) {
                    //如果办理人是当前用户，则置此任务为当前任务
                    if (us.getUserId() == act.getHandler()) {
                        actInst = act;
                        break;
                    }
                }
            }
        }
        //没有我的处理任务，用流程当前任务来
        if (actInst == null) {
            actInst = provider.findTask(procInst.getTaskId());
        }

        return listTask;
    }

    /**
     * 设置单据描述及链接地址，待办中展示
     *
     * @param bill_info
     * @param bill_url
     */
    public void setBillInfo(String bill_code, String bill_info, String bill_url) {
        procInst.setBillInfo(bill_info);
        procInst.setBillCode(bill_code);
        procInst.setUrl(bill_url);
    }

    //启动流程
    public void save() {
        if (opt_mode == FlowConst.OperatorType.NEW) {
            getProcInstDao().insertEntity(procInst);
            getTaskDao().insertEntity(actInst);
            buildBillLog(FlowConst.Button.ADD.value, "-", null, null);
            opt_mode = FlowConst.OperatorType.EDIT;
        } else {
            getProcInstDao().updateEntity(procInst);
            if (actInst != null) {
                getTaskDao().updateEntity(actInst);
            }
        }
    }

    /**
     * 删除流程
     *
     * @throws Exception
     */
    public void delete() {
        getProcInstDao().deleteEntity(procInst);
        getTaskDao().deleteEntity("where tsk_pri_id=?", procInst.getId());
        getCandidateDao().deleteEntity("where cnd_pri_id=?", procInst.getId());
        getCareDao().deleteEntity("where cr_pri_id=?", procInst.getId());
        getCommentDao().deleteEntity("where cmt_pri_id=?", procInst.getId());
        getSignDao().deleteEntity("where sgn_pri_id=?", procInst.getId());
        getTaskRelDao().deleteEntity("where trl_pri_id=?", procInst.getId());
    }

    /**
     * 返回页面前的设置
     */
    public void resetPage() {
        List<Task> listTask = adjustTasks();
        //可选步骤
        tasks.clear();
        String next_text = procDef.getProcInfo().getActNameById(actInst.getActId());
        int index = 0;
        for (Task act : listTask) {
            Map<String, String> row = new HashMap<>();
            tasks.add(row);
            row.put("id", String.valueOf(act.getId()));
            if (act.getId() == actInst.getId() && listTask.size() > 1 && index != 0) {
                next_text = "待" + next_text;
            } else {
                next_text = procDef.getProcInfo().getActNameById(act.getActId());
            }
            row.put("text", next_text);
            index++;
        }
        resetButtons();
    }

    public void resetButtons() {
        Map<String, Integer> buttons = new HashMap<>();
        //按钮
        buttons.clear();
        long user_id = us.getUserId();
        /*
        //菜单权限
        Map<String, String> mapPermisson = MenuHelper.getPermission(context);
        //新增不要显示历史
        boolean isNew = opt_mode == FlowConst.OperatorType.NEW;
        boolean isAdmin = UserBaseEntityBuffer.getInstance().isAdmin(user_id);
        boolean canEdit = MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.ADD.value) || MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.UPD.value);
        //能否新增，菜单有授权+单据有流程分配
        boolean canAdd = canEdit && BillFlowWfHelper.getBillProc(procInst.getBillType(), user_id) != null;
        boolean canDel = MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.DEL.value);
        boolean canDis = MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.DIS.value) && canEdit && !isNew;
        //非制单人能否修改单据
        boolean othCanEdit = BillFlowWfHelper.getPropBool(procDef.getEntityId(), FlowConst.NULL_STR, FlowConst.ProcProperty.CAN_EDIT_OTHER);

        buttons.put(FlowConst.Button.LOG, isNew ? 0 : 1);
        buttons.put(FlowConst.Button.WORD, isNew || !MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.EXP.value) ? 0 : 1);
        buttons.put(FlowConst.Button.EXCEL, isNew || !MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.EXP.value) ? 0 : 1);
        buttons.put(FlowConst.Button.DEL, !isNew ? 1 : 0);
        buttons.put(FlowConst.Button.ADD, canAdd ? 1 : 0);
        //buttons.put(FlowConst.Button.DISUSE, canDis ? 1 : 0);
        buttons.put(FlowConst.Button.DISUSE, 0);

        //单据作废
        if (procInst.getStatu() == FlowConst.InstanceStatu.DISUSE.value) {
            buttons.put(FlowConst.Button.DISUSE, 0);
            buttons.put(FlowConst.Button.DEL, 0);
            context.getDfpResponse().put("buttons", buttons);
            return;
        }
        if (procInst.getStatu() == FlowConst.InstanceStatu.FINISH.value) {
            buttons.put(FlowConst.Button.DEL, 0);
        }
        ActivityEntity act = ActivityEntityBuffer.getInstance().get(actInst.getActId());
        //非制单状态时 编辑权限受 功能权限+流程修改权限+流程节点修改权限
        boolean notMakeEdit = act != null && act.getActEdit() && othCanEdit && MenuHelper.hasPermisson(mapPermisson, FlowConst.MenuFuncRight.UPD.value);
        boolean handler_make = isHandleMakeEx(act);

        //当前任务是我办理的(我的待办任务，前面读取的时候也已经写了办理人为我了)：
        if (act == null || act.getType() == FlowConst.ActivityType.END.value) {
            if (isAdmin) buttons.put(FlowConst.Button.RETAKE, 1);
            buttons.put(FlowConst.Button.DISUSE, 0);
        } else if (user_id.equalsIgnoreCase(actInst.getHandler()) || isAdmin) {
            if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value) {
                //待办任务，显示办理按钮
                if (notMakeEdit || handler_make) {
                    buttons.put(FlowConst.Button.HANDLER, 1);
                    buttons.put(FlowConst.Button.SUBMIT, 0);
                } else {
                    buttons.put(FlowConst.Button.SUBMIT, 1);
                    buttons.put(FlowConst.Button.REJECT, 1);
                    buttons.put(FlowConst.Button.REJECT_MAKE, 1);
                }
            } else if (actInst.getStatu() == FlowConst.ActivityStatu.HANDLE.value) {
                buttons.put(FlowConst.Button.SUBMIT, 1);
                //办理中任务，提交/驳回
                if (actInst.isMake()) {
                    buttons.put(FlowConst.Button.SAVE, 1);
                    if (canDel && !isNew) {
                        buttons.put(FlowConst.Button.DEL, 1);
                    }
                } else {
                    if (notMakeEdit || handler_make) {
                        buttons.put(FlowConst.Button.SAVE, 1);
                    }
                    buttons.put(FlowConst.Button.REJECT, 1);
                    buttons.put(FlowConst.Button.REJECT_MAKE, 1);
                }
            } else if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
                //我已提交的任务，取回(前面已经过滤掉不能取回的任务：下一任务已办理)
                buttons.put(FlowConst.Button.RETAKE, 1);
            }
        } else if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value && new ActinstDao().isMyWaitedTask(actInst.getTaskId(), user_id)) {
            //待办任务，显示办理按钮
            if (notMakeEdit || handler_make) {
                buttons.put(FlowConst.Button.HANDLER, 1);
                buttons.put(FlowConst.Button.SUBMIT, 0);
            } else buttons.put(FlowConst.Button.SUBMIT, 1);
        } else if (actInst.isMake() && actInst.getStatu() == FlowConst.ActivityStatu.HANDLE.value && canEdit && BillFlowWfHelper.getPropBool(procDef.getEntityId(), FlowConst.NULL_STR, FlowConst.ProcProperty.CAN_EDIT_OTHER)) {
            //不是我的任务，制单任务判断属性，是否可以修改他人单据
            ProcdefEntity def = BillFlowWfHelper.getBillProc(procInst.getBillType(), user_id);
            if (def != null && def.getEntityId().equals(procDef.getEntityId())) {
                buttons.put(FlowConst.Button.SAVE, 1);
                buttons.put(FlowConst.Button.SUBMIT, 1);
            }
        }
        context.getDfpResponse().put("buttons", buttons);
         */
    }

    /**
     * 校验权限
     *
     * @throws Exception
     */
    private void checkPermission() {
        if (actInst == null) {
            throw new SwException("没有找到对应的任务步骤！");
        }
        long user_id = us.getUserId();
        //是我办理的单据，正常，返回
        if (actInst.getHandler() <= 0L || user_id == actInst.getHandler()) return;
        //不是我办理的单据，非制单状态，不允许他人操作
        if (!actInst.isMake()) {
            throw new BizException("非制单步骤，必须办理人本人操作！");
        }
    }

    /*public void submitToEnd() {
        long user_id = us.getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
            throw new BizException("当前任务已经提交，不能重复提交");
        }

        if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value && !FlowHelper.isMyWaitedTask(actInst.getEntityId(), user_id)) {
            throw new BizException("您不是当前任务的办理人，不允许提交");
        }
        //校验完毕，接下来干活。
        List<Activity> new_acts = procDef.getProcInfo().findEndActivity();
        if (CommUtil.isEmpty(new_acts)) {
            throw new BizException("结束任务为空，不允许提交到结束！");
        }

        //待新增任务
        List<Task> listInsertTask = new ArrayList<>();
        List<TaskRel> listTaskRel = new ArrayList<>();
        //待新增任务候选人
        List<Candidate> listInsertCand = new ArrayList<>();

        for (Activity act : new_acts) {
            buildTask(act, listInsertTask, listInsertCand, listTaskRel, user_id);
        }

        actInst.setHandler(user_id);
        actInst.setStartTime(DateUtil.nowDateTimeLong());
        actInst.setStatu(FlowConst.ActivityStatu.SUBMIT.value);
        actInst.setRetake(true);

        if (listInsertTask.size() > 0) {
            procInst.setTaskId(listInsertTask.get(0).getEntityId());
        } else {
            procInst.setTaskId(actInst.getEntityId());
        }

        if (new_acts.size() == 1 && new_acts.get(0).getType() == FlowConst.ActivityType.END.value) {
            procInst.setStatu(FlowConst.InstanceStatu.FINISH.value);
            procInst.setEndTime(DateUtil.nowDateTimeLong());
        } else {
            procInst.setStatu(FlowConst.InstanceStatu.RUNING.value);
        }

        //保存入库
        if (!listInsertTask.isEmpty()) {
            getTaskDao().batchInsertEntity(listInsertTask);
            getCandidateDao().batchInsertEntity(listInsertCand);
            getTaskRelDao().batchInsertEntity(listTaskRel);
        }
        getTaskDao().updateEntity(actInst);
        getProcInstDao().updateEntity(procInst);

        buildBillLog(FlowConst.Button.SUBMIT.value, "TODO", actInst, listInsertTask);
        listener.submit(actInst, listInsertTask);
    }*/


    /**
     * 提交
     *
     * @throws Exception
     */
    public void submit() {
        long user_id = us.getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
            throw new BizException("当前任务已经提交，不能重复提交");
        }

        if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value && !FlowHelper.isMyWaitedTask(actInst.getEntityId(), user_id)) {
            throw new BizException("您不是当前任务的办理人，不允许提交");
        }
        //校验完毕，接下来干活。
        List<Activity> new_acts = FlowHelper.getNextActivity(this);

        //待新增任务
        List<Task> listInsertTask = new ArrayList<>();
        List<TaskRel> listTaskRel = new ArrayList<>();
        //待新增任务候选人
        List<Candidate> listInsertCand = new ArrayList<>();

        for (Activity act : new_acts) {
            buildTask(act, listInsertTask, listInsertCand, listTaskRel, user_id);
        }

        actInst.setHandler(user_id);
        actInst.setStartTime(DateUtil.nowDateTimeLong());
        actInst.setStatu(FlowConst.ActivityStatu.SUBMIT.value);
        actInst.setRetake(true);

        if (listInsertTask.size() > 0) {
            procInst.setTaskId(listInsertTask.get(0).getEntityId());
        } else {
            procInst.setTaskId(actInst.getEntityId());
        }

        if (new_acts.size() == 1 && new_acts.get(0).getType() == FlowConst.ActivityType.END.value) {
            procInst.setStatu(FlowConst.InstanceStatu.FINISH.value);
            procInst.setEndTime(DateUtil.nowDateTimeLong());
        } else {
            procInst.setStatu(FlowConst.InstanceStatu.RUNING.value);
        }

        //保存入库
        if (!listInsertTask.isEmpty()) {
            getTaskDao().batchInsertEntity(listInsertTask);
            getCandidateDao().batchInsertEntity(listInsertCand);
            getTaskRelDao().batchInsertEntity(listTaskRel);
        }
        getTaskDao().updateEntity(actInst);
        getProcInstDao().updateEntity(procInst);

        buildBillLog(FlowConst.Button.SUBMIT.value, "TODO", actInst, listInsertTask);
        listener.submit(actInst, listInsertTask);
    }


    public void checkSubmit() {
        long user_id = us.getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
            throw new BizException("当前任务已经提交，不能重复提交");
        }
        if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value && !FlowHelper.isMyWaitedTask(actInst.getEntityId(), user_id)) {
            throw new BizException("您不是当前任务的办理人，不允许提交");
        }
        //comp.checkSubmit(FlowConst.FlowOptType.SUBMIT.value, actInst);
    }

    /**
     * 作废
     *
     * @throws Exception
     */
    public void disuse() {
        long user_id = us.getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (procInst.getStatu() == FlowConst.InstanceStatu.FINISH.value) {
            throw new BizException("当前单据已审批完成，不能作废单据");
        }

        if (actInst.getStatu() == FlowConst.ActivityStatu.WAIT.value && !FlowHelper.isMyWaitedTask(actInst.getEntityId(), user_id)) {
            throw new BizException("您不是当前任务的办理人，不允许作废");
        }

        actInst.setHandler(user_id);
        actInst.setStartTime(DateUtil.nowDateTimeLong());
        actInst.setStatu(FlowConst.ActivityStatu.DISUSE.value);
        procInst.setStatu(FlowConst.InstanceStatu.DISUSE.value);
        procInst.setEndTime(DateUtil.nowDateTimeLong());

        getTaskDao().updateEntity(actInst);
        getProcInstDao().updateEntity(procInst);

        buildBillLog(FlowConst.Button.DISUSE.value, "TODO", actInst, null);
        listener.disuse(actInst);
    }

    /**
     * 提交创建任务
     *
     * @param act
     * @param listInsertTask
     * @param listInsertCand
     * @param listTaskRel
     * @throws Exception
     */
    private void buildTask(Activity act, List<Task> listInsertTask, List<Candidate> listInsertCand, List<TaskRel> listTaskRel, long submit_user) {
        Task task = new Task();
        listInsertTask.add(task);
        task.init();
        task.setEntityId(DbEngine.getInstance().nextId());
        task.setPriId(procInst.getId());
        task.setPriId(procDef.getId());
        task.setActId(act.getId());
        task.setStartTime(DateUtil.nowDateTimeLong());
        task.setStatu(FlowConst.ActivityStatu.WAIT.value);
        task.setMake(false);
        task.setSign(false);
        task.setReject(false);
        task.setRetake(false);
        task.setIsAuto(false);

        TaskRel rel = new TaskRel();
        rel.setEntityId(DbEngine.getInstance().nextId());
        rel.setPriId(task.getPriId());
        rel.setDstTaskId(task.getId());
        rel.setSrcTaskId(actInst.getId());
        listTaskRel.add(rel);

        if (act.getType() != FlowConst.ActivityType.END.value) {
            Set<Long> listHander = new HashSet<>();

            long hanlder_group = act.getHandler().getUserGroup();
            int handler_range = act.getHandler().getFilter();
            //获取所有候选人
            listHander = UserGroupHelper.getUserIdList(hanlder_group, procInst, handler_range, us);

            if (CommUtil.isEmpty(listHander)) {
                throw new BizException("没有符合条件的主办人！" + act.getLabel());
            }
            for (Long user_id : listHander) {
//                if (act.getHandler().getIgnore() && user_id == submit_user) continue;
                Candidate cand = new Candidate();
                cand.init();
                cand.setEntityId(DbEngine.getInstance().nextId());
                cand.setPriId(task.getPriId());
                cand.setTskId(task.getId());
                cand.setUserId(user_id);

                listInsertCand.add(cand);
            }
            if (CommUtil.isEmpty(listInsertCand)) {
                throw new BizException("没有符合条件的主办人！" + act.getLabel());
            }
        }
    }


    /**
     * 取回
     *
     * @throws Exception
     */
    public void retake() {
        long user_id = us.getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() != FlowConst.ActivityStatu.SUBMIT.value) {
            throw new BizException("当前任务非提交状态，不能取回！");
        }
        Activity activityEntity = procDef.getProcInfo().findActivity(actInst.getActId());
        EntityDao<Task> taskDao = getTaskDao();
        if (activityEntity.getType() == FlowConst.ActivityType.END.value) {//从结束取回
            List<Task> new_acts = provider.findBeforeTasks(actInst.getId());
            Task old_act = (Task) actInst.clone();
            old_act.setEntityId(actInst.getEntityId());

            actInst = new_acts.get(0);
            actInst.setHandler(user_id);
            actInst.setStartTime(DateUtil.nowDateTimeLong());
            actInst.setStatu(FlowConst.ActivityStatu.HANDLE.value);
            if (actInst.isMake()) {
                actInst.setHandler(user_id);
            }

            procInst.setTaskId(actInst.getEntityId());
            procInst.setStatu(actInst.isMake() ? FlowConst.InstanceStatu.BEGIN.value: FlowConst.InstanceStatu.RUNING.value);

            taskDao.updateEntity(actInst);
            taskDao.deleteEntity(old_act);
            getTaskRelDao().deleteEntity("where trl_src_task_id=?", old_act.getId());
            getProcInstDao().updateEntity(procInst);
            new_acts.clear();
            new_acts.add(old_act);
            buildBillLog(FlowConst.Button.RETAKE.value, "TODO", actInst, new_acts);
            listener.retake(actInst, new_acts);
//            comp.transCallback(FlowConst.Button.RETAKE.value, actInst, new_acts);
        } else {
            List<Task> new_acts = provider.findAfterTasks(actInst.getId());

            for (Task act : new_acts) {
                if (act.getStatu() != FlowConst.ActivityStatu.WAIT.value) {
                    throw new BizException("后续任务[" + procDef.getProcInfo().getActNameById(act.getActId()) + "]已经办理，不能取回！");
                }
            }

            actInst.setHandler(user_id);
            actInst.setStartTime(DateUtil.nowDateTimeLong());
            actInst.setStatu(FlowConst.ActivityStatu.HANDLE.value);

            procInst.setTaskId(actInst.getEntityId());

            procInst.setStatu(FlowConst.InstanceStatu.RUNING.value);
            if (actInst.isMake()) procInst.setStatu(FlowConst.InstanceStatu.BEGIN.value);

            taskDao.updateEntity(actInst);
            taskDao.batchDelete(new_acts);
            getTaskRelDao().deleteEntity("where trl_src_task_id=?", actInst.getEntityId());
            getProcInstDao().updateEntity(procInst);

            buildBillLog(FlowConst.Button.RETAKE.value, "TODO", actInst, new_acts);
            listener.retake(actInst, new_acts);
        }
    }

    /**
     * 驳回
     *
     * @throws Exception
     */
    public void reject() {
        long user_id = us.getUserId();
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
            throw new BizException("当前任务已提交，不能驳回！");
        }

        //获取前置任务，前置任务需要置为被驳回的办理状态，同时，前置任务的其他后置任务需要删除
        List<Task> new_acts = provider.findBeforeTasks(actInst.getEntityId());
        if (CommUtil.isEmpty(new_acts)) {
            throw new BizException("没有找到前续任务，驳回失败！");
        }
        //记录需要被删除后续任务的任务id
        StringBuilder sqlIn = new StringBuilder(128);
        //设为被驳回状态
        for (Task act : new_acts) {
            act.setStatu(FlowConst.ActivityStatu.HANDLE.value);
            act.setReject(true);

            sqlIn.append(",").append(act.getEntityId());
        }

        actInst.setStatu(FlowConst.ActivityStatu.REJECT.value);

        procInst.setTaskId(new_acts.get(0).getEntityId());
        if (new_acts.size() == 1 && new_acts.get(0).isMake()) {
            procInst.setStatu(FlowConst.InstanceStatu.BEGIN.value);
        } else {
            procInst.setStatu(FlowConst.InstanceStatu.RUNING.value);
        }
        //删除这些任务的后续任务
        List<Task> del_acts = provider.findAfterTasks(sqlIn.substring(1));
        EntityDao<Task> taskDao = getTaskDao();
        taskDao.batchDelete(del_acts);
        taskDao.batchUpdateEntity(new_acts);
        getTaskRelDao().deleteEntity("where trl_src_task_id in (" + sqlIn.substring(1) + ")");
        getProcInstDao().updateEntity(procInst);

        buildBillLog(FlowConst.Button.REJECT.value, "TODO", actInst, new_acts);
        listener.reject(actInst, new_acts);
//        comp.transCallback(FlowConst.FlowOptType.REJECT.value, actInst, new_acts);
    }

    /**
     * 驳回到制单
     *
     * @throws Exception
     */
    public void rejectToMake() {
        //校验权限，当前人员有权限干不，没权限直接抛异常
        checkPermission();
        if (actInst.getStatu() == FlowConst.ActivityStatu.SUBMIT.value) {
            throw new BizException("当前任务已提交，不能驳回！");
        }

        //获取前置任务，前置任务需要置为被驳回的办理状态，同时，前置任务的其他后置任务需要删除
        Task make_act = provider.findMakeTask(actInst.getPriId());
        if (EntityHelper.isEmpty(make_act)) {
            throw new BizException("没有找到制单任务，驳回到制单失败！");
        }

        //设为被驳回状态
        make_act.setStatu(FlowConst.ActivityStatu.HANDLE.value);
        make_act.setReject(true);

        actInst.setStatu(FlowConst.ActivityStatu.REJECT.value);

        procInst.setTaskId(make_act.getEntityId());
        procInst.setStatu(FlowConst.InstanceStatu.BEGIN.value);

        EntityDao<Task> taskDao = getTaskDao();
        taskDao.updateEntity(make_act);
        //删除非制单步骤
        taskDao.deleteEntity("where tsk_pri_id = ? and tsk_id <> ? ", make_act.getPriId(), make_act.getId());
        getTaskRelDao().deleteEntity("where trl_pri_id = ? and trl_tsk_id <> ?", make_act.getPriId(), make_act.getId());
        getProcInstDao().updateEntity(procInst);
        //删除候选人
        getCandidateDao().deleteEntity(" where cnd_pri_id = ? and cnd_tsk_id <> ?", make_act.getPriId(), make_act.getId());

        List<Task> make_acts = new ArrayList<>();
        make_acts.add(make_act);

        buildBillLog(FlowConst.Button.REJECT.value, "TODO", actInst, make_acts);
        listener.reject(actInst, make_acts);
    }

    /**
     * 状态变更，构建日志
     */
    protected void buildBillLog(final String flow_opt, String comment, Task srcTask, List<Task> dstTasks) {
        ProcInfo procInfo = procDef.getProcInfo();

        FlowLog flowLog = new FlowLog();
        flowLog.setComment(comment);
        flowLog.setOpt(flow_opt);

        flowLog.setPriId(procInst.getId());
        flowLog.setTime(DateUtil.nowDateTimeLong());

        flowLog.setUsrId(us.getUserId());
        flowLog.setId(DbEngine.getInstance().nextId());

        String src = null, dst = null;
        StringBuilder sbDst = null;
        if (srcTask != null) {
            src = procInfo.getActNameById(srcTask.getActId());
            flowLog.setActName(src);
            flowLog.setActId(srcTask.getId());
        }
        if (!CommUtil.isEmpty(dstTasks)) {
            sbDst = new StringBuilder();
            for (Task a : dstTasks) {
                sbDst.append("/").append(procInfo.getActNameById(a.getActId()));
            }
            dst = sbDst.substring(1);
        }
        if (FlowConst.Button.RETAKE.value.equals(flow_opt)) {
            flowLog.setInfo(dst + "->" + src);
        } else if (dst != null) {
            flowLog.setInfo(src + "->" + dst);
        } else if (src != null) {
            flowLog.setInfo(src);
        } else {
            flowLog.setInfo("-");
        }
        DbEngine.getInstance().findDao(FlowLog.class).insertEntity(flowLog);

    }

    private EntityDao<ProcInst> getProcInstDao() {
        return DbEngine.getInstance().findDao(ProcInst.class);
    }

    private EntityDao<Task> getTaskDao() {
        return DbEngine.getInstance().findDao(Task.class);
    }

    private EntityDao<Candidate> getCandidateDao() {
        return DbEngine.getInstance().findDao(Candidate.class);
    }

    private EntityDao<Care> getCareDao() {
        return DbEngine.getInstance().findDao(Care.class);
    }

    private EntityDao<Comment> getCommentDao() {
        return DbEngine.getInstance().findDao(Comment.class);
    }

    private EntityDao<Sign> getSignDao() {
        return DbEngine.getInstance().findDao(Sign.class);
    }

    private EntityDao<TaskRel> getTaskRelDao() {
        return DbEngine.getInstance().findDao(TaskRel.class);
    }
}
